#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"

#define MAX_PRINT 20

// Reset border values to T_edge
void set_edge_values(double *grid, int Nx, int Ny, double T_edge) {

  for (int i=0; i<Nx; i++) {
    grid[     0*Nx+i] = T_edge;
    grid[(Ny-1)*Nx+i] = T_edge;
  }

  for (int j=0; j<Ny; j++) {
    grid[j*Nx+0] = T_edge;
    grid[j*Nx+(Nx-1)] = T_edge;
  }
}

// Update grid by averaging adjacent points
void update_stencil(double *grid_old, double *grid_new, int Nx, int j_start, int j_end) {

    for (int j = j_start; j < j_end; j++) {
      for (int i = 1; i < Nx-1; i++) {
        grid_new[j*Nx+i] = 0.25*(grid_old[j*Nx+i+1]   + grid_old[j*Nx+i-1]+
                                 grid_old[(j+1)*Nx+i] + grid_old[(j-1)*Nx+i]);
      }
    }
}

// Find max difference between values on two grids
double find_max_diff(double *grid_old, double *grid_new, int Nx, int j_start, int j_end) {

    double max_diff = 0.0;
    for (int j = j_start; j < j_end; j++) {
      for (int i = 1; i < Nx-1; i++) {
        double diff = fabs((double) (grid_old[j*Nx+i] - grid_new[j*Nx+i]));
        max_diff = (max_diff < diff) ? diff : max_diff;
      }
    }
    return max_diff;
}

typedef struct {
  unsigned Nx, Ny, max_iters;
  double T_edge;
  double convergence_threshold;
} input_parameters;

int main(int argc, char* argv[]) {
  MPI_Init(&argc, &argv);

  int rank, size;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  MPI_Barrier(MPI_COMM_WORLD);

  /* INPUT PROCESSING */

  input_parameters p;
  if (rank == 0) {
    printf("\n COMP4300/8300 Prac Hot Plate Program\n");
    printf(" Number of processors %d\n", size);
    printf(" Input number of X and Y data points \n");
    if (scanf("%d%d", &p.Nx, &p.Ny) != 2) {
      fprintf(stderr, "error: invalid Nx, Ny\n");
      exit(1);
    }
    printf(" Input temperatures along edges \n");
    if (scanf("%lf", &p.T_edge) != 1) {
      fprintf(stderr, "error: invalid temperature\n");
      exit(1);
    }
    printf(" Input maximum iterations \n");
    if (scanf("%d", &p.max_iters) != 1) {
      fprintf(stderr, "error: invalid X, Y\n");
      exit(1);
    }
    printf(" Input convergence \n");
    if (scanf("%lf", &p.convergence_threshold) != 1) {
      fprintf(stderr, "error: invalid X, Y\n");
      exit(1);
    }

    for (int i=0; i<72; i++) printf("-");
    printf("\n");
    printf(" Nx                    %8d \n", p.Nx);
    printf(" Ny                    %8d \n", p.Ny);
    printf(" T_edge                %8.2lf\n", p.T_edge);
    printf(" max_iters             %8d\n", p.max_iters);
    printf(" convergence_threshold %8.2lf\n", p.convergence_threshold);
    for (int i=0; i<72; i++) printf("-");
    printf("\n");
  }

  MPI_Bcast(&p, sizeof(p), MPI_BYTE, 0, MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);

/***********************************************************************/
/*                     MAIN COMPUTE PART                               */
/***********************************************************************/
  // Allocate data arrays
  double* grid_old = (double*) malloc(p.Nx*p.Ny*sizeof(double));
  double* grid_new = (double*) malloc(p.Nx*p.Ny*sizeof(double));

  /* *******Parallel Solution*********/
  if (rank == 0) printf("\n Begin Parallel Iterations\n");
  for (int i = 0; i < p.Nx*p.Ny; i++) { 
    grid_old[i] = 0.0; grid_new[i] = 0.0; 
  }

  set_edge_values(grid_old, p.Nx, p.Ny, p.T_edge);

  MPI_Barrier(MPI_COMM_WORLD);
  double t_start = MPI_Wtime();

  const int chunk_size = (p.Ny-2+size-1)/size;
  const int j_start = rank*chunk_size+1;
  const int j_end   = (j_start + chunk_size) > p.Ny-1 ? p.Ny-1 : (j_start + chunk_size);

  for (int iter = 1; iter <= p.max_iters; iter++) {
    // update local values
    update_stencil(grid_old, grid_new, p.Nx, j_start, j_end);

/*#define ORIG*/
/*#define BLOCKING_DEADLOCK_FREE*/
#define NON_BLOCKING

#ifdef ORIG
    // Send right-most column to rank+1
    if (rank+1 < size) {
      MPI_Send(&grid_new[(j_end-1)*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 2, MPI_COMM_WORLD);
    }
    if (rank-1 >= 0) {
      MPI_Recv(&grid_new[(j_start-1)*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 2, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
    }

    // Send left-most column to rank-1
    if (rank-1 >= 0) {
      MPI_Send(&grid_new[j_start*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 1, MPI_COMM_WORLD);
    }
    if (rank+1 < size) {
      MPI_Recv(&grid_new[j_end*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 1, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
    }
#endif
#ifdef BLOCKING_DEADLOCK_FREE
    // Send to rank+1
    if (rank+1 < size && rank % 2 == 0) {
      MPI_Send(&grid_new[(j_end-1)*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 2, MPI_COMM_WORLD);
    }
    if (rank-1 >= 0) {
      MPI_Recv(&grid_new[(j_start-1)*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 2, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
    }
    if (rank+1 < size && rank % 2 != 0) {
      MPI_Send(&grid_new[(j_end-1)*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 2, MPI_COMM_WORLD);
    }

    // Send to rank-1
    if (rank-1 >= 0 && rank % 2 == 0) {
      MPI_Send(&grid_new[j_start*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 1, MPI_COMM_WORLD);
    }
    if (rank+1 < size) {
      MPI_Recv(&grid_new[j_end*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 1, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
    }
    if (rank-1 >= 0 && rank % 2 != 0) {
      MPI_Send(&grid_new[j_start*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 1, MPI_COMM_WORLD);
    }
#endif
#ifdef NON_BLOCKING
    MPI_Request request[4];
    int num_requests = 0; // count of requests made
    // Send to rank+1
    if (rank+1 < size) {
      MPI_Isend(&grid_new[(j_end-1)*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 2,
          MPI_COMM_WORLD, &request[num_requests++]);
      MPI_Irecv(&grid_new[j_end*p.Nx], p.Nx, MPI_DOUBLE, rank+1, 1, MPI_COMM_WORLD,
               &request[num_requests++]);
    }
    if (rank-1 >= 0) {
      MPI_Isend(&grid_new[j_start*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 1, MPI_COMM_WORLD,
               &request[num_requests++]);
      MPI_Irecv(&grid_new[(j_start-1)*p.Nx], p.Nx, MPI_DOUBLE, rank-1, 2, MPI_COMM_WORLD,
               &request[num_requests++]);
    }
    MPI_Waitall(num_requests, request, MPI_STATUSES_IGNORE);
#endif

    // Fix boundaries in grid_new
    set_edge_values(grid_new, p.Nx, p.Ny, p.T_edge);

    // Find the maximum difference in absolute value among all entries of two 
    // consecutive iterates for the locally owned set of grid points (i.e., exclude ghost points).
    // This is required to evaluate the early termination criterion.
    double max_difference;
    double local_max_difference = find_max_diff(grid_old, grid_new, p.Nx, j_start, j_end);
    MPI_Allreduce(&local_max_difference, &max_difference, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

    if (rank == 0) 
      printf(" iteration %d convergence %lf\n", iter, max_difference);
    
    if (max_difference < p.convergence_threshold)
      break;

    // Copy grid for next iteration
    for (int i = 0; i < p.Nx*p.Ny; i++) grid_old[i] = grid_new[i];
  }

  MPI_Barrier(MPI_COMM_WORLD);
  double t_par = MPI_Wtime() - t_start;
  if (rank == 0)
    printf("Total time parallel   %8.4lf\n", t_par);

  /* Printout the result if small!*/
  if (p.Nx <= MAX_PRINT && p.Ny <= MAX_PRINT && size == 1) {
    printf(" final grid values");
    for (int j = 0; j < p.Ny; j++) {
      printf("\n Col %2d:   ", j);
      for (int i = 0; i < p.Nx; i++) {
        printf("%10.2e", grid_new[j*p.Nx+i]);
        if (i%5 == 4) printf("\n           ");
      }
    }
  }
  if (rank == 0)
    printf("\n");

/***********************************************************************/
/*                    End MAIN COMPUTE PART                            */
/***********************************************************************/
  free(grid_old); free(grid_new);  

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();

  return 0;
} //main()
